#!/usr/bin/pl -s
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Yet Another NIM 1.0
% Anicka Bernathova (anicka@anicka.net) wrote me in January 2007
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generally useful predicates working mostly with lists

%returns length of a list
%len(+List,-Number)
len([],0).
len([_|X],N):- len(X,N1), N is N1+1.

%concatenates two lists
%cat(+List,+List,-List)
cat([],A,A).
cat([A|X],B,[A|Z]):-cat(X,B,Z).

%returns min member of a list
%min(+List,-Number)
min([A|X],B):-min([A|X],B,A).
min([],B,B).
min([A|X],B,C):-(A<C,!,min(X,B,A);min(X,B,C)).

%trims zeros from a list
%trim(+List,-List)
trim([],[]).
trim([A|X],Y):-A=:=0, trim(X,Y).
trim([A|X],[A|Y]):-A=\=0, trim(X,Y).

%trims a list of positions
%clean(+MoveList,-CleanMoveList)
clean([],[]).
clean([X|XS],[A|L]):-trim(X,A), clean(XS,L).

%sorts every position in a list of positions
%sortl(+MoveList,-SortedMoveList)
sortl([],[]).
sortl([X|XS],[A|L]):-msort(X,A), sortl(XS,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Following predicates are used for a basic variation of NIM

%calculates xor of a position
%getxor(+List,-Number)
getxor(Input,Xor):-getxor2(0,Input,Xor).
getxor2(Temp,[],Temp).
getxor2(Temp,[A|X],Xor):-Temp2 is xor(Temp,A), getxor2(Temp2,X,Xor).

%getout(+List,-List,+Xor)
getout([A|X],[B|Y],Xor):- B is xor(A,Xor), B<A, Y=X. 
getout([A|X],[A|Y],Xor):- getout(X,Y,Xor). 

%gets xor to 0
%get_0(+List,-List)
get_0(Input,Output):- getxor(Input,Xor), getout(Input,Output,Xor).

%returns random move
%get_random(+List,-List)
get_random(A,B):- len(A,C), D is random(C)+1, getmoven(D,A,B).
getmoven(N,[A|X],[B|Y]):- A>=N, B is A-N, Y=X.
getmoven(N,[A|X],[A|Y]):- A<N, C is N-A, getmoven(C,X,Y). 

%returns a move, either winning (if possible) or a random one
%play(+List,-List)
play(A,B):-getxor(A,Xor), plej2(A,Xor,BB), !, trim(BB,B).
plej2(A,Xor,B):-Xor=:=0, get_random(A,B).
plej2(A,Xor,B):-Xor=\=0, getout(A,B,Xor).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicates used for variations where we know no better solution
% than a backtracking

%returns a position reachable within the standard rules of NIM
movestd([A|L], [B|L]) :- AA is A-1, between(0,AA,B).	% B>=0, B<A
movestd([A|L], [A|M]) :- movestd(L,M).

%returns a position reachable within an extra rule: 
%removing a one match from every heap is allowed
moveext1([], []).
moveext1([A|L], [B|M]) :- A>0, B is A-1, moveext1(L,M).

%returns a position reachable within an extra rule:
%removing of the same number of matches from one heap is allowed
moveext2([],[]).
moveext2([A|L], [B|N]):-min([A|L],M), between(1,M,S), B is A-S, moveext3(L,N,S).
moveext3([],[],_).
moveext3([A|L],[B|M],S):-B is A-S, moveext3(L,M,S).

%returns a legal move, either a standard or an extra one
%notice: right extra move is determined during a game
move([],_) :- !, fail.
move(A,B) :- movestd(A,B) ; moveext(A,B).

%returns all legal moves
moves(A,B) :- bagof(X,move(A,X),B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is a real backtracking part. We are using a cache to
% determine status of a position faster. See developer's documentation 
% for more information.
% This is how a tree looks like: s(L, R, V)

%ins1(+Tree1,+Key,+Value,-OldValue,-Tree1)
ins1([], K, V, W, T) :- ins1(s([],[],[]),K,V,W,T).
ins1(s(L,R,W), 0, V, W, s(L,R,V)) :- !.
ins1(s(L,R,Z), X, V, W, s(LL,R,Z)) :- X mod 2 =:= 0, X2 is X//2, ins1(L,X2,V,W,LL).
ins1(s(L,R,Z), X, V, W, s(L,RR,Z)) :- X mod 2 =:= 1, X2 is X//2, ins1(R,X2,V,W,RR).

%inserts a value of the KeyList into the tree
%ins(+Tree,+KeyList,+Value,-Tree)
ins(T,[],V,TT) :- ins1(T,0,V,_,TT).
ins(T,[K|L],V,TT) :- ins1(T,K,SS,S,TT), ins(S,L,V,SS).

%find1(+Tree1,+Key,-Value)
find1(T,K,V) :- ins1(T,K,_,V,_).

%finds a value of the KeyList
%find(+Tree,+KeyList,-Value)
find(T,[],V):-find1(T,0,V).
find(T,[K|KL],V):-find1(T,K,V1),find(V1,KL,V).

%determines whether a position is won
%valc(+Tree,+Position,-Value,-Tree)
valc(T,P,V,T2):-find(T,P,W), (W==[],!,val(T,P,V,T1),ins(T1,P,V,T2);V=W,T=T2).
val(T,P,V,T2):-P\=[],moves(P,R), clean(R,O),sortl(O,Q),val2(T,Q,V,T2).
val(T,[],V,T2):-val2(T,[],V,T2).
val2(T,[],0,T).
val2(T,[X|Q],V,T2):-valc(T,X,W,T1),(W=0,!,T2=T1,V=1;val2(T1,Q,V,T2)).                                

%returns a move, winning if possible
%hintsc(+Tree,+Position,-Position,-Tree)
hintsc(T,L,L2,T2):-(move(L,LL), trim(LL,L2),msort(L2,L3),valc(T,L3,0,T2),!; moves(L,[L2|_])).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% game - following predicates take care for interaction with user

%starts a game
game:-init, askfortype(C), askforlist([A|B]), !, abolish(moveext,2), go_which([A|B],C).

%determines which loop should be chosen
%and sets moveext according to the loop
go_which(A,2):-!, AA=.. [moveext,K,L], BB=..[moveext1,K,L], X=(AA:-BB), assert(X), go(A,[],0).
go_which(A,3):-!, AA=.. [moveext,K,L], BB=..[moveext2,K,L], X=(AA:-BB), assert(X), go(A,[],1).
go_which(A,_):- go(A).

%asks user which type of a game he wishes to play
askfortype(A):-write('Enter a type of a game you would like to play:\n'),
               write('1. for a traditional NIM (this is default), \n'),
               write('2. to be allowed to remove one match from all the heaps besides standard rules \n'),
               write('3. to be allowed to remove any constant number of matches from all the heaps besides standard rules.\n'), 
               read(A).

%here comes a stuff necessary for a game initialization. Nothing really important, in fact :-)
init:-write('Ahoj babi! '),write('\n').

%tests whether the list provided by user is a valid NIM position
valid([]).
valid([A|AX]):-A>=0, valid(AX).

%asks user for an initial position
askforlist(A):-write('Enter an initial position, list of heaps:'), read(B), write('\n'),( (B==[];(\+ valid(B))), !, write('Argh... Try it once again! \n'), askforlist(A) ; A=B ).

%determines whether we won or game continues
return_move([]):-write([]), write(' :-)'), !, fail.
return_move(A):-write(A).

%determines whether user won or game continues
test_state([]):-write(' :-('), !, fail.
test_state(_).

%plays user's move if legal or fail
%remove(+Number,+Number,+List,-List)
remove(_,Y,_,_) :- Y=<0, !, write('This really does not look like a legal move!\n'), fail.
remove(X,Y,[A|T],[B|T]) :- X=:=1, ( A>=Y, B is A-Y ;
				    A<Y,  !, write('There is not that many matches!\n'), fail).
remove(X,Y,[A|T],[A|S]) :- X=\=1, X1 is X-1, remove(X1,Y,T,S).
remove(_,_,[],_) :- write('There is not that many heaps!\n'), fail.

% and once again for moves containing more heaps 
%remove2(+Number,+List,-List)
remove2(Y,B,_):-min(B,M), Y>M, !, write('There is not that many matches!\n'), fail.
remove2(_,[],[]).
remove2(Y,[A|B],[K|L]):-K is A-Y, remove2(Y,B,L).

%game loop. If you are a spaghetti lover, you will surely appreciate this part
%go(+Position) - loop used for a basic variation
go(A):-trim(A,T),
       test_state(T), 
       write('List of heaps after my move: '),
       play(T,B),
       trim(B,K), 
       return_move(K), 
       write('\n'), 
       write('Enter a number of a heap you want to reduce: '), 
       read(X),
       write('Enter number of matches you want to drop: '), 
       read(Y),
       remove(X,Y,B,G),
       write('List of heaps after your move: '), write(G), write('\n'), 
       go(G). 
%...and advanced variations
%go(+Position,+GameType)
go(A,Tree,P):-trim(A,T),
         test_state(T),
         write('List of heaps after my move: '),
         hintsc(Tree,T,K,Tree2),
         trim(K,B),
         return_move(B),
         write('\n'),
         write('Enter a number of a heap you want to reduce, 0. to reduce all the heaps: '),
         read(X),
         ((P=\=0 ;  X=\=0), !, write('Enter number of matches you want to drop: '),
         read(Y); Y=1),
         (X=:=0, !,
          remove2(Y,B,G) ;
          remove(X,Y,B,G)),
         write('List of heaps after your move: '), write(G), write('\n'),
         go(G,Tree2,P).                  
